import os
import torch

from diffusers import DDIMScheduler
from diffusers.loaders.lora import LoraLoaderMixin

from image_retrieval_configs import parse_args_and_update_config
from pipeline_gen_mi_1phase import MISelectionPipeline



def main():
    if 'load config':
        config = parse_args_and_update_config()
   
    if 'load sd & dora':
        model_id = config.pretrained_model_name_or_path

        scheduler = DDIMScheduler.from_pretrained(model_id, subfolder="scheduler")
        
        if 'spright' not in model_id:
            pipe = MISelectionPipeline.from_pretrained(model_id, scheduler=scheduler, torch_dtype=torch.float32)
        else:
            pipe = MISelectionPipeline.from_pretrained(model_id, scheduler=scheduler,  torch_dtype=torch.float32,
                use_safetensors=True,
                token=os.getenv("HF_TOKEN"),
            )
        
        if config.dora_ckpt_dir:
   
            state_dict, network_alphas = pipe.lora_state_dict(
                config.dora_ckpt_dir,
                unet_config=pipe.unet.config,
            )

            is_correct_format = all("lora" in key or "dora_scale" in key for key in state_dict.keys())
            if not is_correct_format:
                raise ValueError("Invalid DoRA checkpoint.")
        
            #####################################################################################

            LoraLoaderMixin.load_lora_into_unet(
                state_dict = state_dict,
                network_alphas = network_alphas, 
                unet = pipe.unet,
            )

            print('load DoRA weights done!')
        else:
            print('no DoRA weights loaded!')
        
        pipe = pipe.to("cuda")

        pipe.mi_weight_init()

    if 'load prompts':
        
        if config.prompt_file_name is not None:
            with open(config.prompt_file_name, "r") as f:
                prompts = f.read().splitlines()
                prompts = [d.strip("\n").split("\t")[0] for d in prompts] 
            print(f"prompts from {config.prompt_file_name}:  len={len(prompts)}") 
        elif config.prompt:
            prompts = [config.prompt]
        else:
            raise ValueError("No prompts input")

    
    if "makedir":
        seed = config.seed
        data_outdir = config.outdir
        if not os.path.exists(data_outdir):
            os.makedirs(data_outdir)
    
    num_images_per_prompt = config.num_images_per_prompt
    z_sample_num = config.z_sample_num

    for prompt in prompts:
        if not os.path.isfile(os.path.join(data_outdir, f"{prompt}.png")):
            
            with torch.no_grad(), torch.cuda.amp.autocast():

                print(f'prompt {prompt}, seed {seed}')
                generator = torch.Generator(device="cuda").manual_seed(seed)
    
                images = pipe(prompt = prompt, 
                              num_images_per_prompt = num_images_per_prompt,
                              num_inference_steps=z_sample_num, 
                              generator=generator,
                              data_outdir = data_outdir,
             
                            ).images
    
   
if __name__ == "__main__":
    main()
