import copy
import gc
import logging
import math
import os
from pathlib import Path

import numpy as np
import torch
import torch.nn.functional as F
import torch.utils.checkpoint
import transformers
from accelerate import Accelerator
from accelerate.logging import get_logger
from accelerate.utils import ProjectConfiguration, set_seed
from peft import LoraConfig
from peft.utils import get_peft_model_state_dict, set_peft_model_state_dict
from PIL import Image
from torchvision import transforms
from tqdm.auto import tqdm
from transformers import AutoTokenizer

from termcolor import colored
from transformers import CLIPTextModel

# clip
import open_clip
from sentence_transformers import util as clip_util
# dataset
import pandas as pd
from datasets import load_dataset
from datasets import Features
from datasets import Value
import random
import wandb
import diffusers
from diffusers import (
    AutoencoderKL,
    DDPMScheduler,
    DiffusionPipeline,
    DPMSolverMultistepScheduler,
    UNet2DConditionModel,
)
from diffusers.loaders import LoraLoaderMixin
from diffusers.utils import (
    check_min_version,
    convert_state_dict_to_diffusers,
    convert_unet_state_dict_to_peft,
)
check_min_version("0.28.0.dev0")
from diffusers.utils.torch_utils import is_compiled_module

from finetuning_configs import parse_args_finetuning

###################################################################################################################

logger = get_logger(__name__)

def log_validation(
    pipeline, args, accelerator,
    ):
    logger.info( f"Running validation..." )

    # We train on the simplified learning. If we were previously predicting a variance, we need the scheduler to ignore it.
    scheduler_args = {}
    if "variance_type" in pipeline.scheduler.config:
        variance_type = pipeline.scheduler.config.variance_type
        if variance_type in ["learned", "learned_range"]:
            variance_type = "fixed_small"
        scheduler_args["variance_type"] = variance_type
    pipeline.scheduler = DPMSolverMultistepScheduler.from_config(pipeline.scheduler.config, **scheduler_args)
    pipeline = pipeline.to(accelerator.device)
    pipeline.set_progress_bar_config(disable=True)

    
    generator = torch.Generator(device=accelerator.device).manual_seed(args.seed) if args.seed else None
    all_images = []
    all_prompts = []
    for prompt in args.validation_prompts:
        with torch.autocast(device_type='cuda'):
            images = pipeline(prompt, 
                        num_images_per_prompt = args.num_validation_images, 
                        num_inference_steps=30,
                        generator=generator).images 
            all_images.extend(images)
            all_prompts.extend([prompt] * args.num_validation_images)

    phase_name = "validation"
    for tracker in accelerator.trackers:
        if tracker.name == "wandb":
            tracker.log(
                {
                    phase_name: [
                        wandb.Image(all_images[i], caption=f"{i}: {all_prompts[i]}") 
                        for i in range(len(all_images))
                    ]
                }
            )

    del pipeline
    gc.collect()
    torch.cuda.empty_cache()


def set_logger(args, accelerator, logger):
    # Make one log on every process with the configuration for debugging.
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO,
    )
    logger.info(accelerator.state, main_process_only=False)
    if accelerator.is_local_main_process:
        transformers.utils.logging.set_verbosity_warning()
        diffusers.utils.logging.set_verbosity_info()
    else:
        transformers.utils.logging.set_verbosity_error()
        diffusers.utils.logging.set_verbosity_error()
    
    # Handle the repository creation 
    if accelerator.is_main_process:
        if args.outdir is not None:
            os.makedirs(args.outdir, exist_ok=True)
    
        # Add log file 
        formatter = logging.Formatter('%(asctime)s %(filename)s %(levelname)s %(message)s')
        fh = logging.FileHandler(os.path.join(args.outdir, 'log.txt'), mode='a', encoding='utf8')
        fh.setFormatter(formatter)
        logger.logger.addHandler(fh)
        logger.info(args)

def main(args):
    # set logger
    accelerator_project_config = ProjectConfiguration(project_dir=args.outdir, 
                                                      total_limit=args.checkpoints_total_limit,)
    accelerator = Accelerator(
        gradient_accumulation_steps=args.gradient_accumulation_steps,
        mixed_precision=None,
        log_with="wandb",
        project_config=accelerator_project_config,
    )
    set_logger(args, accelerator, logger)
    
    # set the training seed
    if args.seed is not None:
        set_seed(args.seed)
    
  
    # load SD components
    tokenizer = AutoTokenizer.from_pretrained(args.pretrained_model_name_or_path, subfolder="tokenizer", use_fast=False)
    noise_scheduler = DDPMScheduler.from_pretrained(args.pretrained_model_name_or_path, subfolder="scheduler")
    text_encoder = CLIPTextModel.from_pretrained(args.pretrained_model_name_or_path, subfolder="text_encoder")
    vae = AutoencoderKL.from_pretrained(args.pretrained_model_name_or_path, subfolder="vae", )
    unet = UNet2DConditionModel.from_pretrained(args.pretrained_model_name_or_path, subfolder="unet" )
    # We only train the additional adapter DoRA layers
    vae.requires_grad_(False)
    text_encoder.requires_grad_(False)
    unet.requires_grad_(False)
    # Move unet, vae and text_encoder to device and cast to weight_dtype
    weight_dtype = torch.float32
    unet.to(accelerator.device, dtype=weight_dtype)
    vae.to(accelerator.device, dtype=weight_dtype)
    text_encoder.to(accelerator.device, dtype=weight_dtype)

    # add new DoRA weights to the attention layers
    unet_lora_config = LoraConfig(r=args.rank,
                                  lora_alpha=int(args.rank * args.alpha_div_rank),
                                  use_dora=args.use_dora,
                                  init_lora_weights="gaussian", target_modules=["to_k", "to_q", "to_v", "to_out.0", "add_k_proj", "add_v_proj"], 
                                 )
    unet.add_adapter(unet_lora_config) 

    # create custom saving & loading hooks so that `accelerator.save_state(...)` serializes in a nice format
    def unwrap_model(model):
        model = accelerator.unwrap_model(model)
        model = model._orig_mod if is_compiled_module(model) else model
        return model
    
    def save_model_hook(models, weights, outdir):
        if accelerator.is_main_process:
            unet_lora_layers_to_save = None
            for model in models:

                if isinstance(model, type(unwrap_model(unet))):
                    unet_lora_layers_to_save = convert_state_dict_to_diffusers(
                        get_peft_model_state_dict(model))
                else:
                    raise ValueError(f"unexpected save model: {model.__class__}")
                
                # make sure to pop weight so that corresponding model is not saved again
                weights.pop()

            LoraLoaderMixin.save_lora_weights(
                outdir,
                unet_lora_layers=unet_lora_layers_to_save,
            )

    def load_model_hook(models, input_dir):
        unet_ = None
        while len(models) > 0:
            model = models.pop()

            if isinstance(model, type(unwrap_model(unet))):
                unet_ = model
            else:
                raise ValueError(f"unexpected save model: {model.__class__}")

        lora_state_dict, network_alphas = LoraLoaderMixin.lora_state_dict(input_dir)

        unet_state_dict = {f'{k.replace("unet.", "")}': v for k, v in lora_state_dict.items() if k.startswith("unet.")}
        unet_state_dict = convert_unet_state_dict_to_peft(unet_state_dict)
        incompatible_keys = set_peft_model_state_dict(unet_, unet_state_dict, adapter_name="default")

        if incompatible_keys is not None:
            # check only for unexpected keys
            unexpected_keys = getattr(incompatible_keys, "unexpected_keys", None)
            if unexpected_keys:
                logger.warning(
                    f"Loading adapter weights from state_dict led to unexpected keys not found in the model: "
                    f" {unexpected_keys}. "
                )

    accelerator.register_save_state_pre_hook(save_model_hook)
    accelerator.register_load_state_pre_hook(load_model_hook)


    # Optimizer creation
    params_to_optimize_unet = list(filter(lambda p: p.requires_grad, unet.parameters()))
    params_to_optimize = params_to_optimize_unet
    optimizer_class = torch.optim.AdamW
    optimizer = optimizer_class(
        params_to_optimize,
        lr=args.learning_rate,
        betas=(args.adam_beta1, args.adam_beta2),
        weight_decay=args.adam_weight_decay,
        eps=args.adam_epsilon,
    )

    # Dataset creation
    data_paths = []
    prompts = []

    for path, subdirs, files in os.walk(args.train_data_dir): 
       
        print('training data folder: ', path)
        
        for name in files:
            data_path = os.path.join(path, name)
            prompt = data_path.split('/')[-1].split('.png')[0]

            if '-' in prompt:
                prompt = prompt.split('-')[0]

            data_paths.append(data_path)
            prompts.append(prompt)

    if (args.max_train_samples is not None) and (len(data_paths) > args.max_train_samples):  
        print(f'prompts: select {args.max_train_samples} out of {len(data_paths)}')
        data_paths, prompts = data_paths[:args.max_train_samples], prompts[:args.max_train_samples]
    print(f'image paths: {len(data_paths)}, prompts: {len(prompts)}')

    df_csv_path = args.outdir + '/train_data.csv'
    df_column_names = ['img_path', 'prompt']
    image_column, caption_column = df_column_names[0], df_column_names[1]

    df = pd.DataFrame(columns = df_column_names)
    df['img_path'] = data_paths 
    df['prompt'] = prompts
    df.to_csv(df_csv_path)  

    data_files = {}
    data_files["train"] = df_csv_path
    dataset = load_dataset(
        "csv",
        data_files= data_files,
        delimiter= ',',
        column_names= df_column_names, 
        features= Features({'img_path': Value('string'), 'prompt': Value('string')}),
        skiprows= 1
        )


    # Preprocessing the datasets.

    # tokenize input captions
    def tokenize_captions(examples, is_train=True):
        captions = []
        for caption in examples[caption_column]:
            if isinstance(caption, str):
                captions.append(caption)
            elif isinstance(caption, (list, np.ndarray)):
                # take a random caption if there are multiple
                captions.append(random.choice(caption) if is_train else caption[0])
            else:
                raise ValueError(
                    f"Caption column `{caption_column}` should contain either strings or lists of strings."
                )
        inputs = tokenizer(
            captions, max_length=tokenizer.model_max_length, padding="max_length", truncation=True, return_tensors="pt"
        )
        return inputs.input_ids

    # transform the images
    train_transforms = transforms.Compose(
        [
            transforms.Resize(args.resolution, interpolation=transforms.InterpolationMode.BILINEAR),
            transforms.CenterCrop(args.resolution) if args.center_crop else transforms.RandomCrop(args.resolution),
            transforms.ToTensor(),
            transforms.Normalize([0.5], [0.5]),
        ]
    )
    
    # combine the 2 above
    def preprocess_train(examples):
        images = [Image.open(image).convert("RGB") for image in examples[image_column]]
        examples["pixel_values"] = [train_transforms(image) for image in images]
        examples["input_ids"]    = tokenize_captions(examples)
        return examples
    # set the training transforms
    with accelerator.main_process_first():
        train_dataset = dataset["train"].with_transform(preprocess_train)


    # DataLoaders creation
    def collate_fn(examples):
        pixel_values = torch.stack([example["pixel_values"] for example in examples]).to(memory_format=torch.contiguous_format).float()
        input_ids    = torch.stack([example["input_ids"]    for example in examples])
        return {"pixel_values": pixel_values, "input_ids": input_ids}
    
    train_dataloader = torch.utils.data.DataLoader(
        train_dataset,
        shuffle=True,
        collate_fn=collate_fn,
        batch_size=args.train_batch_size,
        num_workers=args.dataloader_num_workers,
    )


    unet, optimizer, train_dataloader = accelerator.prepare( unet, optimizer, train_dataloader ) 


    # Scheduler and math around the number of training steps.
    num_update_steps_per_epoch = math.ceil(len(train_dataloader) / args.gradient_accumulation_steps)
    # Afterwards we recalculate our number of training epochs
    args.num_train_epochs      = math.ceil(args.max_train_steps  / num_update_steps_per_epoch)
    total_batch_size = args.train_batch_size * accelerator.num_processes * args.gradient_accumulation_steps

    # We need to initialize the trackers we use, and also store our configuration.
    # The trackers initializes automatically on the main process.
    if accelerator.is_main_process:
        tracker_config = vars(copy.deepcopy(args))
        tracker_config.pop('validation_prompts')
        accelerator.init_trackers(args.outdir.split('/')[-1], config=tracker_config)

    # Train!
    logger.info("***** Running training *****")
    logger.info(f"  Num examples = {len(train_dataset)}")
    logger.info(f"  Num batches each epoch = {len(train_dataloader)}")
    logger.info(f"  Num Epochs = {args.num_train_epochs}")
    logger.info(f"  Instantaneous batch size per device = {args.train_batch_size}")
    logger.info(f"  Total train batch size (w. parallel, distributed & accumulation) = {total_batch_size}")
    logger.info(f"  Gradient Accumulation steps = {args.gradient_accumulation_steps}")
    logger.info(f"  Total optimization steps = {args.max_train_steps}")
    
    global_step = 0
    first_epoch = 0
    optimizer.zero_grad()
    
    progress_bar = tqdm(
        range(0, args.max_train_steps),
        initial=global_step,
        desc="Steps",
        disable=not accelerator.is_local_main_process,  # Only show the progress bar once on each machine.
    )

    for epoch in range(first_epoch, args.num_train_epochs):
        unet.train()
        for step, batch in enumerate(train_dataloader):
            with accelerator.accumulate(unet):
      
                # Convert images to latent space
                model_input = vae.encode(batch["pixel_values"].to(dtype=weight_dtype)).latent_dist.sample()
                model_input = model_input * vae.config.scaling_factor

                # Sample noise that we'll add to the latents
                noise = torch.randn_like(model_input)
                bsz, channels, height, width = model_input.shape

                # Sample a random timestep for each image
                randint_start = int(noise_scheduler.config.num_train_timesteps/2) if args.earlyts else 0
                timesteps = torch.randint(
                                        randint_start, noise_scheduler.config.num_train_timesteps, 
                                        (bsz,), device=model_input.device
                                        ).long()
              
                # Add noise to the model input according to the noise magnitude at each timestep
                # (this is the forward diffusion process)
                noisy_model_input = noise_scheduler.add_noise(model_input, noise, timesteps)

                # Get the text embedding for conditioning
                encoder_hidden_states = text_encoder(batch["input_ids"], return_dict=False)[0]

                # Predict the noise residual
                model_pred = unet(
                    noisy_model_input,
                    timesteps,
                    encoder_hidden_states,
                    class_labels=None,
                    return_dict=False,
                )[0]

                
                target = noise

                loss = F.mse_loss(model_pred.float(), target.float(), reduction="mean")

                accelerator.backward(loss)
                
                if accelerator.sync_gradients:
                    accelerator.clip_grad_norm_(params_to_optimize, args.max_grad_norm)
                
                optimizer.step()
              
                optimizer.zero_grad()

            # Checks if the accelerator has performed an optimization step behind the scenes
            if accelerator.sync_gradients:
                progress_bar.update(1)
                global_step += 1
                acc_log_dict = {"train_loss": loss.detach().item(), "lr": optimizer.param_groups[0]['lr']} 

                if accelerator.is_main_process and global_step % args.validation_steps == 0: 
                        
                    # create pipeline
                    pipeline = DiffusionPipeline.from_pretrained(
                        args.pretrained_model_name_or_path,
                        unet=unwrap_model(unet),
                        torch_dtype=weight_dtype,
                    )
                    log_validation(
                        pipeline, args, accelerator,
                    )
    
                    save_path = os.path.join(args.outdir, f"checkpoint-{global_step}")
                    accelerator.save_state(save_path)
                    logger.info(f"Saved state to {save_path}")
                    
                    print( colored(f"Saved state to {save_path}",  'green') )
            

                progress_bar.set_postfix(**acc_log_dict)
                accelerator.log(acc_log_dict, step=global_step)

                if (global_step >= args.max_train_steps):
                    save_path = os.path.join(args.outdir, f"checkpoint-{global_step}-last")
                    accelerator.save_state(save_path)
                    logger.info(f"Saved state to {save_path}")
                    break


    accelerator.end_training()


if __name__ == "__main__":
    args = parse_args_finetuning()
    main(args)


